try:
    import cv2
except ImportError:
    cv2 = None


def annotate_image(image, detection):
    """ Annotate images with object detection results

    # Arguments:
        image: numpy array representing the image used for detection
        detection: `DetectionResult` result from SKIL on the same image

    # Return value:
        annotated image as numpy array
    """
    if cv2 is None:
        raise Exception("OpenCV is not installed.")
    objects = detection.get('objects')
    if objects:
        for detect in objects:
            confs = detect.get('confidences')
            max_conf = max(confs)
            max_index = confs.index(max_conf)
            classes = detect.get('predictedClasses')
            max_class = classes[max_index]
            h = detect.get('height')
            w = detect.get('width')
            center_x = detect.get('centerX')
            center_y = detect.get('centerY')
            lower = (int(center_x - w/2), int(center_y - h/2))
            upper = (int(center_x + w/2), int(center_y + h/2))
            im = cv2.rectangle(image, lower, upper, (255, 0, 0), 3)
            cv2.putText(image, max_class + " " + str(max_conf),
                        upper, cv2.FONT_HERSHEY_SIMPLEX, 1, 0, 2)
    return image



from skil_client import AddCredentialsRequest
import skil_client


class Credentials:
    """Credentials

    SKIL resource credentials manage cloud provider and other credentials for you.
    Currently supported credentials are AWS, Azure, GoogleCloud and Hadoop

    # Arguments:
        skil: `Skil` server instance
        cred_type: credentials type string, either "AWS", "Azure", GoogleCloud" or "Hadoop"
        uri: URI pointing to the credentials
        name: Name of the resource
        create   
    """

    def __init__(self, skil, cred_type, uri, name=None, create=True, *args, **kwargs):
        self.skil = skil
        if not cred_type in ["AWS", "Azure", "GoogleCloud", "Hadoop"]:
            raise ValueError("cred_type {} not supported".format(cred_type))
        self.cred_type = cred_type
        self.uri = uri
        self.name = name

        if create:
            request_body = AddCredentialsRequest(
                type=self.cred_type, name=self.name, uri=self.uri)
            response = self.skil.api.add_credentials(request_body)
            self.id = response.credential_id

    def delete(self):
        """Delete these credentials from SKIL.
        """
        if self.id:
            self.skil.api.delete_credentials_by_id(self.id)
            self.id = None


def AWS(Credentials):
    def __init__(self, skil, uri, name=None):
        super(AWS, self).__init__(
            skil=skil, cred_type="AWS", uri=uri, name=name)


def Azure(Credentials):
    def __init__(self, skil, uri, name=None):
        super(Azure, self).__init__(
            skil=skil, cred_type="Azure", uri=uri, name=name)


def GoogleCloud(Credentials):
    def __init__(self, skil, uri, name=None):
        super(GoogleCloud, self).__init__(
            skil=skil, cred_type="GoogleCloud", uri=uri, name=name)


def Hadoop(Credentials):
    def __init__(self, skil, uri, name=None):
        super(Hadoop, self).__init__(
            skil=skil, cred_type="Hadoop", uri=uri, name=name)


def delete_credentials_by_id(skil, credentials_id):
    try:
        resp = skil.api.delete_credentials_by_id(credentials_id)
    except:
        raise Exception(
            'Credentials with id {} could not be deleted, make sure they exist.'.format(credentials_id))


def get_credentials_by_id(skil, credentials_id):
    response = skil.api.get_credentials_by_id(credentials_id)
    cred = Credentials(skil=skil, cred_type=response.type, uri=response.uri,
                       name=response.name, create=False)
    cred.id = credentials_id
    return cred



import skil_client


class JobConfiguration:
    """JobConfiguration

    A SKIL job configuration collects all data needed to set up and run a SKIL Job.
    SKIL currently has inference and training jobs, each come with their respective
    configuration.

    # Arguments:
        skil_model: a `skil.Model` instance
        compute_resource: `skil.resources.compute.ComputeResource' instance, created before running a job.
        storage_resource: `skil.resources.storage.StorageResource` instance created before runnning a job.
        output_path: string with path to folder in which job output should be stored.
        data_set_provider_class: name of the class to be used as `DataSetProvider` in SKIL
        is_multi_data_set: boolean, whether data set uses `MultiDataSet` interface.
        verbose: boolean, log level. Set True for detailed logging.

    """

    # TODO: provide a smart default for output_path relative to input data or model path.

    def __init__(self, skil_model, compute_resource, storage_resource,
                 output_path, data_set_provider_class,
                 is_multi_data_set, verbose):
        self.model = skil_model
        self.compute_id = compute_resource.resource_id
        self.storage_id = storage_resource.resource_id
        self.output_path = output_path
        self.dsp = data_set_provider_class
        self.mds = is_multi_data_set
        self.verbose = verbose


class InferenceJobConfiguration(JobConfiguration):
    """InferenceJobConfiguration

    Configuration for a SKIL inference job. On top of what you need to specify for a base JobConfiguration,
    you need to set the batch size for the model as well.

    # Arguments:
        skil_model: a `skil.Model` instance
        batch_size: int, data batch size to run inference with on the model.
        compute_resource: `skil.resources.compute.ComputeResource' instance, created before running a job.
        storage_resource: `skil.resources.storage.StorageResource` instance created before runnning a job
        output_path: string with path to folder in which job output should be stored.
        data_set_provider_class: name of the class to be used as `DataSetProvider` in SKIL
        is_multi_data_set: boolean, whether data set uses `MultiDataSet` interface.
        verbose: boolean, log level. Set True for detailed logging.

    """
    # TODO signature to aim for: (model, data_path (what), data_format (how), storage_id (where), compute_id)
    # TODO: data format may eventually even be inferred automatically. no reason not to. we can do so for models.

    # TODO: we could even consider *setting* compute and storage resources for workspaces
    #       or experiments. No need to specify this every time.

    # TODO batch_size should either be known to the model or the data, I don't believe this.
    # TODO KILL DSP!!!1ONE
    # TODO There must be a way to hide is_multi_data_set

    def __init__(self, skil_model, batch_size, compute_resource, storage_resource, output_path,
                 data_set_provider_class,
                 is_multi_data_set=False,
                 verbose=False):
        super(InferenceJobConfiguration, self).__init__(
            skil_model, compute_resource,
            storage_resource, output_path, data_set_provider_class,
            is_multi_data_set, verbose)

        self.batch_size = batch_size


class TrainingJobConfiguration(JobConfiguration):

    """TrainingJobConfiguration

    Configuration for a SKIL training job. On top of what you need to specify for a base JobConfiguration,
    you need to set the number of epochs to train for, a (distributed) training configuration and provide
    information about how to evaluate your model.

    # Arguments:
        skil_model: a `skil.Model` instance
        compute_resource: `skil.resources.compute.ComputeResource' instance, created before running a job.
        storage_resource: `skil.resources.storage.StorageResource` instance created before runnning a job
        output_path: string with path to folder in which job output should be stored.
        data_set_provider_class: name of the class to be used as `DataSetProvider` in SKIL
        is_multi_data_set: boolean, whether data set uses `MultiDataSet` interface.
        verbose: boolean, log level. Set True for detailed logging.
    """
    # TODO signature to aim for: (model, num_epochs, data_path, eval_data_path, eval_types,
    #  data_format, storage_id, compute_id)
    # TODO what if we want to split data on the go. what about validation data? cross validation?
    #  current concept seems insufficient to cover this properly.
    # TODO model_path, config_path, model config path... why both? should be able to guess this
    # TODO model_history_url=None, can we infer this from experiment?
    # TODO model_history_id=None, this is the workspace id
    # TODO model_instance_id=None, is this alternatively to model/config path? don't get it.
    # TODO: eval_type make this a proper class instead of (or additionally to) strings.
    # TODO: allow multiple eval metrics?!
    # TODO: the training master config should be deconstructed. maybe provide this to the job.run(...) as argument.
    # TODO: user should just be *handed* a ui, not take care of a $%$%! URL.

    def __init__(self,  skil_model, num_epochs,
                 eval_type,
                 eval_data_set_provider_class,  # good lord
                 compute_resource, storage_resource,
                 output_path,
                 data_set_provider_class,
                 is_multi_data_set=False,
                 ui_url=None,
                 verbose=False):
        super(TrainingJobConfiguration, self).__init__(
            skil_model, compute_resource,
            storage_resource, output_path, data_set_provider_class,
            is_multi_data_set, verbose)

        self.num_epochs = num_epochs
        self.eval_dsp = eval_data_set_provider_class
        self.eval_type = eval_type
        self.ui_url = ui_url


class TrainingJob:
    """TrainingJob

    Initialize and run a SKIL training job. If a distributed config is provided,
    SKIL will run your model on Spark. Otherwise it will carry out regular training
    on provided resources.

    # Arguments:
        training_config: `TrainingJobConfiguration` instance
        distributed_config: `DistributedConfiguration` instance

    """

    def __init__(self, skil, training_config, distributed_config=None):

        self.skil = skil
        self.training_config = training_config
        self.tm = distributed_config

        training_create_job_request = skil_client.CreateJobRequest(
            compute_resource_id=self.training_config.compute_id,
            storage_resource_id=self.training_config.storage_id,
            job_args=self._training_job_args(),
            output_file_name=self.training_config.output_path
        )

        # TODO: why do we need to specify the training type here if the request already knows it?
        self.skil.api.create_job("TRAINING", training_create_job_request)

    def run(self):
        pass

    def _training_job_args(self):
        tc = self.training_config
        tm = self.tm

        inference = "-i false "
        output = "-o {} ".format(tc.output_path)
        num_epochs = "--batchSize {} ".format(tc.num_epochs)
        model_path = "-mo {} ".format(tc.model.model_path)
        dsp = "-dsp {} ".format(tc.dsp)
        eval_dsp = "--evalDataSetProviderClass {} ".format(tc.eval_dsp)
        eval_type = "--evalType {} ".format(tc.eval_type)
        tm = "-tm {} ".format(tc.tm)
        mds = "--multiDataSet {} ".format(_bool_to_string(tc.mds))
        verbose = "--verbose {} ".format(_bool_to_string(tc.verbose))

        return inference + output + num_epochs + model_path + dsp + \
            eval_dsp + eval_type + tm + mds + verbose


class InferenceJob:

    def __init__(self, skil, inference_config):

        self.skil = skil
        self.inference_config = inference_config

        inference_create_job_request = skil_client.CreateJobRequest(
            compute_resource_id=self.inference_config.compute_id,
            storage_resource_id=self.inference_config.storage_id,
            job_args=self._inference_job_args(),
            output_file_name=self.inference_config.output_path
        )

        self.skil.api.create_job("INFERENCE", inference_create_job_request)

    def run(self):
        pass

    def _inference_job_args(self):
        ic = self.inference_config

        inference = "-i true "
        output = "-o {} ".format(ic.output_path)
        batch_size = "--batchSize {} ".format(ic.batch_size)
        model_path = "-mo {} ".format(ic.model.model_path)
        dsp = "-dsp {} ".format(ic.dsp)
        mds = "--multiDataSet {} ".format(_bool_to_string(ic.mds))
        verbose = "--verbose {} ".format(_bool_to_string(ic.verbose))

        return inference + output + batch_size + model_path + dsp + \
            mds + verbose


def get_job_by_id(skil_server, job_id):
    # TODO is this available?
    pass


def _bool_to_string(bool):
    return "true" if bool else "false"



import skil_client
import time
import uuid
import numpy as np
import requests
import json
import os

try:
    import cv2
except ImportError:
    cv2 = None


class Service:
    """Service

    A service is a deployed model.

    # Arguments:
        skil: `Skil` server instance
        model: `skil.Model` instance
        deployment: `skil.Deployment` instance
        model_deployment: result of `deploy_model` API call of a model
    """

    def __init__(self, skil, model, deployment, model_deployment):
        self.skil = skil
        self.model = model
        self.model_name = self.model.name
        self.model_deployment = model_deployment
        self.deployment = deployment

    def start(self):
        """Starts the service.
        """
        if not self.model_deployment:
            self.skil.printer.pprint(
                "No model deployed yet, call 'deploy()' on a model first.")
        else:
            self.skil.api.model_state_change(
                self.deployment.id,
                self.model_deployment.id,
                skil_client.SetState("start")
            )

            self.skil.printer.pprint(">>> Starting to serve model...")
            while True:
                time.sleep(5)
                model_state = self.skil.api.model_state_change(
                    self.deployment.id,
                    self.model_deployment.id,
                    skil_client.SetState("start")
                ).state
                if model_state == "started":
                    time.sleep(15)
                    self.skil.printer.pprint(
                        ">>> Model server started successfully!")
                    break
                else:
                    self.skil.printer.pprint(">>> Waiting for deployment...")

    def stop(self):
        """Stops the service.
        """
        # TODO: test this
        self.skil.api.model_state_change(
            self.deployment.id,
            self.model_deployment.id,
            skil_client.SetState("stop")
        )

    @staticmethod
    def _indarray(np_array):
        """Convert a numpy array to `skil_client.INDArray` instance.

        # Arguments
            np_array: `numpy.ndarray` instance.

        # Returns
            `skil_client.INDArray` instance.
        """
        return skil_client.INDArray(
            ordering='c',
            shape=list(np_array.shape),
            data=np_array.reshape(-1).tolist()
        )

    def predict(self, data):
        """Predict for given batch of data.

        # Arguments:
            data: `numpy.ndarray` (or list thereof). Batch of input data, or list of batches for multi-input model.

        # Returns
            `numpy.ndarray` instance for single output model and list of `numpy.ndarray` for multi-output model.
        """
        if isinstance(data, list):
            inputs = [self._indarray(x) for x in data]
        else:
            inputs = [self._indarray(data)]

        classification_response = self.skil.api.multipredict(
            deployment_name=self.deployment.name,
            model_name=self.model_name,
            version_name="default",
            body=skil_client.MultiPredictRequest(
                id=str(uuid.uuid1()),
                needs_pre_processing=False,
                inputs=inputs
            )
        )
        outputs = classification_response.outputs
        outputs = [np.asarray(o.data).reshape(o.shape) for o in outputs]
        if len(outputs) == 1:
            return outputs[0]
        return outputs

    def predict_single(self, data):
        """Predict for a single input.

        # Arguments:
            data: `numpy.ndarray` (or list thereof). Input data.

        # Returns
            `numpy.ndarray` instance for single output model and list of `numpy.ndarray` for multi-output model.
        """
        if isinstance(data, list):
            inputs = [self._indarray(np.expand_dims(x, 0)) for x in data]
        else:
            inputs = [self._indarray(np.expand_dims(data, 0))]

        classification_response = self.skil.api.multipredict(
            deployment_name=self.deployment.name,
            model_name=self.model_name,
            version_name="default",
            body=skil_client.MultiPredictRequest(
                id=str(uuid.uuid1()),
                needs_pre_processing=False,
                inputs=inputs
            )
        )
        output = classification_response.outputs[0]
        return np.asarray(output.data).reshape(output.shape)

    def detect_objects(self, image, threshold=0.5, needs_preprocessing=False, temp_path='temp.jpg'):
        """Detect objects in an image for this service. Only works when deploying an object detection
            model like YOLO or SSD.

        # Argments
            image: `numpy.ndarray`. Input image to detect objects from.
            threshold: floating point between 0 and 1. bounding box threshold, only objects with at
                least this threshold get returned.
            needs_preprocessing: boolean. whether input data needs pre-processing
            temp_path: local path to which intermediate numpy arrays get stored.

        # Returns
            `DetectionResult`, a Python dictionary with labels, confidences and locations of bounding boxes
                of detected objects.
        """
        if cv2 is None:
            raise Exception("OpenCV is not installed.")
        cv2.imwrite(temp_path, image)
        url = 'http://{}/endpoints/{}/model/{}/v{}/detectobjects'.format(
            self.skil.config.host,
            self.model.deployment.name,
            self.model.name,
            self.model.version
        )

        # TODO: use the official "detectobject" client API, once fixed in skil_client
        # print(">>>> TEST")
        # resp = self.skil.api.detectobjects(
        #     id='foo',
        #     needs_preprocessing=False,
        #     threshold='0.5',
        #     image_file=temp_path,
        #     deployment_name=self.model.deployment.name,
        #     version_name='default',
        #     model_name=self.model.name
        # )

        with open(temp_path, 'rb') as data:
            resp = requests.post(
                url=url,
                headers=self.skil.auth_headers,
                files={
                    'file': (temp_path, data, 'image/jpeg')
                },
                data={
                    'id': self.model.id,
                    'needs_preprocessing': 'true' if needs_preprocessing else 'false',
                    'threshold': str(threshold)
                }
            )
        if os.path.isfile(temp_path):
            os.remove(temp_path)

        return json.loads(resp.content)



import skil_client
from .base import Skil
import uuid
import json


class Deployment:
    """Deployments operate independently of workspaces to ensure that there are
    no accidental interruptions or mistakes in a production environment.

    # Arguments:
        skil: `Skil` server instance.
        name: string. Name for the deployment.
        id: Unique id for the deployment. If `None`, a unique id will be generated.
    """
    # TODO: starting from skil 1.2 deployments are linked to workspaces and experiments.
    # Make sure to keep this up-to-date.

    def __init__(self, skil=None, name=None, deployment_id=None):
        if not skil:
            skil = Skil()  # TODO: take care of auth
        if deployment_id is not None:
            response = skil.api.deployment_get(deployment_id)
            if response is None:
                raise KeyError('Deployment not found: ' + str(deployment_id))
            self.response = response
            self.name = self.response.name
            self.id = deployment_id
        else:
            self.name = name if name else 'deployment-' + str(uuid.uuid1())[:8]
            create_deployment_request = skil_client.CreateDeploymentRequest(
                self.name)
            self.response = skil.api.deployment_create(
                create_deployment_request)
            self.id = self.response.id

    def get_config(self):
        return {
            'deployment_id': self.id
        }

    def save(self, file_name):
        config = self.get_config()
        with open(file_name, 'w') as f:
            json.dump(config, f)

    @classmethod
    def load(cls, file_name):
        with open(file_name, 'r') as f:
            config = json.load(f)
        skil = Skil.from_config()
        return get_deployment_by_id(skil, config['deployment_id'])


def get_deployment_by_id(skil, deployment_id):
    """ Get model deployment by ID

    # Arguments
        skil: `Skil` server instance
        deployment_id: deployment ID
    """
    return Deployment(skil=skil, deployment_id=deployment_id)



import json


class DistributedConfiguration:

    def __init__(self):
        self.config = None

    def to_json(self):
        return json.dumps(self.config)


class ParameterAveraging(DistributedConfiguration):
    def __init__(self, num_workers, batch_size, averaging_frequency=5,
                 num_batches_prefetch=0, collect_stats=False):
        """ParameterAveraging

        # Arguments:
            num_workers: number of Spark workers/executors.
            batch_size: batch size used for model training
            averaging_frequency: int, after how many batches of training averaging takes place
            num_batches_prefetch: int, how many batches to pre-fetch, deactivated if 0.
            collect_stats: boolean, if statistics get collected during training
         """
        self.config = {
            'numWorkers': num_workers,
            'batchSize': batch_size,
            'averagingFrequency': averaging_frequency,
            'numBatchesPrefetch': num_batches_prefetch,
            'collectStats': collect_stats
        }


class ParameterSharing(DistributedConfiguration):
    def __init__(self, num_workers, batch_size,
                 shake_frequency=0, min_threshold=1e-5, update_threshold=1e-3, workers_per_node=-1,
                 num_batches_prefetch=0, step_delay=50, step_trigger=0.05, threshold_step=1e-5,
                 collect_stats=False):
        """ParameterSharing
            num_workers: number of Spark workers/executors.
            batch_size: batch size used for model training
            shake_frequency: shake frequency
            min_threshold: minimum threshold
            update_threshold: update threshold
            workers_per_node: workers per node
            num_batches_prefetch: number of batches to prefetch
            step_delay: step delay
            step_trigger: step trigger
            threshold_step: threshold step
            collect_stats: boolean, if statistics get collected during training
        """
        self.config = {
            'numWorkers': num_workers,
            'batchSize': batch_size,
            'shakeFrequency': shake_frequency,
            'minThreshold': min_threshold,
            'updateThreshold': update_threshold,
            'workersPerNode': workers_per_node,
            'numBatchesPrefetch': num_batches_prefetch,
            'stepDelay': step_delay,
            'thresholdStep': threshold_step,
            'collectStats': collect_stats
        }



class Resource:
    """Resource

    A SKIL  resource is an abstraction for (cloud and on-premise)
    compute or storage capabilities.
    """
    __metaclass__ = type

    def __init__(self, skil, resource_id=None, *args, **kwargs):
        """Add the resource to SKIL.
        """
        self.skil = skil
        self.resource_id = resource_id

    def delete(self):
        """Delete the resource from SKIL.
        """
        if self.resource_id:
            self.skil.api.delete_resource_by_id(resource_id=self.resource_id)



import skil_client
from .base import Resource


class AzureStorage(Resource):
    """AzureStorage

    SKIL Azure storage resource.

    # Arguments:
        skil: `Skil` server instance
        name: Resource name
        container_name: Azure storage container name
        resource_id: optional resource ID to retrieve an existing resource
        create: boolean, for internal use only. whether to create a new resource or not
    """

    def __init__(self, skil, name, container_name, credential_uri,
                 resource_id=None, create=True):

        super(AzureStorage, self).__init__(skil)

        self.name = name
        self.container_name = container_name
        self.credential_uri = credential_uri
        self.resource_id = resource_id

        if create:
            resource_response = self.skil.api.add_resource(skil_client.AddResourceRequest(
                resource_name=self.name,
                resource_details=skil_client.AzureStorageResourceDetails(
                    container_name=self.container_name
                ),
                credential_uri=self.credential_uri,
                type="STORAGE",
                sub_type="AzureStorage")
            )
            self.resource_id = resource_response.get("resourceId")
        else:
            if resource_id is None:
                raise ValueError(
                    'If create is False you need to provide a valid resource_id')


class GoogleStorage(Resource):
    """GoogleStorage

    SKIL Google storage resource.

    # Arguments:
        skil: `Skil` server instance
        name: Resource name
        project_id: Google project ID
        bucket_name: bucket name
        resource_id: optional resource ID to retrieve an existing resource
        create: boolean, for internal use only. whether to create a new resource or not
    """

    def __init__(self, skil, name, project_id, bucket_name, credential_uri,
                 resource_id=None, create=True):

        super(GoogleStorage, self).__init__(skil)
        self.name = name
        self.project_id = project_id
        self.bucket_name = bucket_name
        self.credential_uri = credential_uri
        self.resource_id = resource_id

        if create:
            resource_response = self.skil.api.add_resource(skil_client.AddResourceRequest(
                resource_name=self.name,
                resource_details=skil_client.GoogleStorageResourceDetails(
                    project_id=self.project_id,
                    bucket_name=self.bucket_name
                ),
                credential_uri=self.credential_uri,
                type="STORAGE",
                sub_type="GoogleStorage")
            )
            self.resource_id = resource_response.get("resourceId")
        else:
            if resource_id is None:
                raise ValueError(
                    'If create is False you need to provide a valid resource_id')


class HDFS(Resource):
    """HDFS

    SKIL HDFS resource.

    # Arguments:
        skil: `Skil` server instance
        name: Resource name
        name_node_host: host of the name node
        name_node_port: port of the name node
        resource_id: optional resource ID to retrieve an existing resource
        create: boolean, for internal use only. whether to create a new resource or not
    """

    def __init__(self, skil, name, name_node_host, name_node_port, credential_uri,
                 resource_id=None, create=True):

        super(HDFS, self).__init__(skil)
        self.name = name
        self.name_node_host = name_node_host
        self.name_node_port = name_node_port
        self.credential_uri = credential_uri
        self.resource_id = resource_id

        if create:
            resource_response = self.skil.api.add_resource(skil_client.AddResourceRequest(
                resource_name=self.name,
                resource_details=skil_client.HDFSResourceDetails(
                    name_node_host=self.name_node_host,
                    name_node_port=self.name_node_port
                ),
                credential_uri=self.credential_uri,
                type="STORAGE",
                sub_type="HDFS")
            )
            self.resource_id = resource_response.get("resourceId")
        else:
            if resource_id is None:
                raise ValueError(
                    'If create is False you need to provide a valid resource_id')


class S3(Resource):
    """S3

    SKIL S3 resource.

    # Arguments:
        skil: `Skil` server instance
        name: Resource name
        bucket: S3 bucket name
        region: AWS region
        resource_id: optional resource ID to retrieve an existing resource
        create: boolean, for internal use only. whether to create a new resource or not
    """

    def __init__(self, skil, name, bucket, region, credential_uri,
                 resource_id=None, create=True):

        super(S3, self).__init__(skil)
        self.name = name
        self.bucket = bucket
        self.region = region
        self.credential_uri = credential_uri
        self.resource_id = resource_id

        if create:
            resource_response = self.skil.api.add_resource(skil_client.AddResourceRequest(
                resource_name=self.name,
                resource_details=skil_client.S3ResourceDetails(
                    bucket=self.bucket,
                    region=self.region
                ),
                credential_uri=self.credential_uri,
                type="STORAGE",
                sub_type="S3")
            )
            self.resource_id = resource_response.get("resourceId")
        else:
            if resource_id is None:
                raise ValueError(
                    'If create is False you need to provide a valid resource_id')



import skil_client
from .base import Resource


class EMR(Resource):
    """EMR

    AWS Elastic Map Reduce compute resource

    # Arguments:
        skil: `Skil` server instance
        name: Name of the resource
        region: AWS region of the EMR cluster
        credential_uri: path to credential file
        cluster_id: ID of the EMR cluster
        resource_id: optional resource ID to retrieve an existing resource
        create: boolean, for internal use only. whether to create a new resource or not
    """
    # TODO: if cluster_id is None, spin up a cluster and retrieve id (requires work in SKIL core)
    # TODO: can we hide setting credentials? i.e. can these be put into a
    #   little config file (similar to what we do in pydl4j?).

    def __init__(self, skil, name, region, credential_uri, cluster_id=None,
                 resource_id=None, create=True):

        super(EMR, self).__init__(skil)
        self.name = name
        self.region = region
        self.credential_uri = credential_uri
        self.cluster_id = cluster_id
        self.resource_id = resource_id

        if create:
            resource_response = self.skil.api.add_resource(skil_client.AddResourceRequest(
                resource_name=self.name,
                resource_details=skil_client.EMRResourceDetails(
                    cluster_id=self.cluster_id,
                    region=self.region
                ),
                credential_uri=self.credential_uri,
                type="COMPUTE",
                sub_type="EMR")
            )
            self.resource_id = resource_response.get("resourceId")
        else:
            if resource_id is None:
                raise ValueError(
                    'If create is False you need to provide a valid resource_id')


class DataProc(Resource):
    """DataProc

    Google cloud engine DataProc compute resource

    # Arguments:
        skil: `Skil` server instance
        name: Resource name
        project_id: GCE project ID
        region: GCE region
        cluster_name: DataProc cluster name
        resource_id: optional resource ID to retrieve an existing resource
        create: boolean, for internal use only. whether to create a new resource or not
    """

    def __init__(self, skil, name, project_id, region, spark_cluster_name, credential_uri,
                 resource_id=None, create=True):
        super(DataProc, self).__init__(skil)
        self.name = name
        self.project_id = project_id
        self.region = region
        self.credential_uri = credential_uri
        self.cluster_name = spark_cluster_name
        self.resource_id = resource_id
        if create:
            resource_response = self.skil.api.add_resource(skil_client.AddResourceRequest(
                resource_name=self.name,
                resource_details=skil_client.DataProcResourceDetails(
                    project_id=self.project_id,
                    region=self.region,
                    spark_cluster_name=self.cluster_name
                ),
                credential_uri=self.credential_uri,
                type="COMPUTE",
                sub_type="DataProc")
            )
            self.resource_id = resource_response.get("resourceId")
        else:
            if resource_id is None:
                raise ValueError(
                    'If create is False you need to provide a valid resource_id')


class HDInsight(Resource):
    """HDInsight

    Azure HDInsight compute resource.

    # Arguments:
        skil: `Skil` server instance
        name: Resource name
        subscription_id: Azure subscription ID
        resource_group_name: Resource group name  # TODO: is this SKIL or Azure?
        cluster_name: HDInsight cluster name
        resource_id: optional resource ID to retrieve an existing resource
        create: boolean, for internal use only. whether to create a new resource or not
    """

    def __init__(self, skil, name, subscription_id, resource_group_name, cluster_name, credential_uri,
                 resource_id=None, create=True):
        super(HDInsight, self).__init__(skil)
        self.name = name
        self.subscription_id = subscription_id
        self.resource_group_name = resource_group_name
        self.cluster_name = cluster_name
        self.credential_uri = credential_uri
        self.resource_id = resource_id

        if create:
            resource_response = self.skil.api.add_resource(skil_client.AddResourceRequest(
                resource_name=self.name,
                resource_details=skil_client.HDInsightResourceDetails(
                    subscription_id=self.subscription_id,
                    resource_group_name=self.resource_group_name,
                    cluster_name=self.cluster_name
                ),
                credential_uri=self.credential_uri,
                type="COMPUTE",
                sub_type="HDInsight")
            )
            self.resource_id = resource_response.get("resourceId")
        else:
            if resource_id is None:
                raise ValueError(
                    'If create is False you need to provide a valid resource_id')


class YARN(Resource):
    """YARN

    YARN compute resource for local Spark computation on YARN.

    # Arguments:
        skil: `Skil` server instance
        name: Resource name
        local_spark_home: full path to local Spark binary
        resource_id: optional resource ID to retrieve an existing resource
        create: boolean, for internal use only. whether to create a new resource or not
    """

    def __init__(self, skil, name, local_spark_home, credential_uri,
                 resource_id=None, create=True):
        super(YARN, self).__init__(skil)
        self.name = name
        self.local_spark_home = local_spark_home
        self.credential_uri = credential_uri
        self.resource_id = resource_id

        if create:
            resource_response = self.skil.api.add_resource(skil_client.AddResourceRequest(
                resource_name=self.name,
                resource_details=skil_client.YARNResourceDetails(
                    local_spark_home=self.local_spark_home
                ),
                credential_uri=self.credential_uri,
                type="COMPUTE",
                sub_type="YARN")
            )
            self.resource_id = resource_response.get("resourceId")
        else:
            if resource_id is None:
                raise ValueError(
                    'If create is False you need to provide a valid resource_id')



import skil_client
from skil_client.rest import ApiException as api_exception
import uuid
import json
from .base import Skil
from .workspaces import get_workspace_by_id, WorkSpace


class Experiment:
    """Experiments in SKIL are useful for defining different model configurations, 
    encapsulating training of models, and carrying out different data cleaning tasks.

    Experiments have a one-to-one relationship with Notebooks and have their own 
    storage mechanism for saving different model configurations when seeking a best 
    candidate.

    # Arguments:
        work_space: `WorkSpace` instance. If `None` a workspace will be created.
        experiment_id: integer. Unique id for workspace. If `None`, a unique id will be generated.
        name: string. Name for the experiment.
        description: string. Description for the experiment.
        verbose: boolean. If `True`, api response will be printed.
        create: boolean. If `True` a new experiment will be created.
    """

    def __init__(self, work_space=None, experiment_id=None, name='experiment',
                 description='experiment', verbose=False, create=True,
                 *args, **kwargs):
        if create:
            if not work_space:
                self.skil = Skil.from_config()
                work_space = WorkSpace(self.skil)
            self.work_space = work_space
            self.skil = self.work_space.skil
            self.id = experiment_id if experiment_id else work_space.id + \
                "_experiment_" + str(uuid.uuid1())
            self.name = name
            experiment_entity = skil_client.ExperimentEntity(
                experiment_id=self.id,
                experiment_name=name,
                experiment_description=description,
                model_history_id=self.work_space.id
            )

            add_experiment_response = self.skil.api.add_experiment(
                self.skil.server_id,
                experiment_entity
            )
            self.experiment_entity = experiment_entity

            if verbose:
                self.skil.printer.pprint(add_experiment_response)
        else:
            experiment_entity = work_space.skil.api.get_experiment(
                work_space.skil.server_id,
                experiment_id
            )
            self.experiment_entity = experiment_entity
            self.work_space = work_space
            self.id = experiment_id
            self.name = experiment_entity.experiment_name

    def get_config(self):
        return {
            'experiment_id': self.id,
            'experiment_name': self.name,
            'workspace_id': self.work_space.id
        }

    def save(self, file_name):
        config = self.get_config()
        with open(file_name, 'w') as f:
            json.dump(config, f)

    @classmethod
    def load(cls, file_name):
        with open(file_name, 'r') as f:
            config = json.load(f)

        skil_server = Skil()
        work_space = get_workspace_by_id(skil_server, config['workspace_id'])
        return get_experiment_by_id(work_space, config['experiment_id'])

    def delete(self):
        """Deletes the experiment.
        """
        try:
            api_response = self.skil.api.delete_experiment(
                self.work_space.id, self.id)
            self.skil.printer.pprint(api_response)
        except api_exception as e:
            self.skil.printer.pprint(
                ">>> Exception when calling delete_experiment: %s\n" % e)


def get_experiment_by_id(work_space, experiment_id):
    return Experiment(work_space=work_space, experiment_id=experiment_id, create=False)



from.common import get_resource_by_id


class ResourceGroup:
    '''ResourceGroup

    SKIL resource groups can be used to group skil.resources.base.Resource instances
    into logical groups. You first create a group and then add resources to the group
    later on, using `add_resource`.

    # Arguments:
        skil: `Skil` server instance
        group_name: Name of the resource group
        group_id: optional group ID to retrieve an existing resource group
        create: boolean, for internal use only. whether to create a new resource or not
    '''

    def __init__(self, skil, group_name, group_id=None, create=True):
        self.skil = skil
        self.group_name = group_name
        self.group_id = group_id

        if create:
            resp = self.skil.api.add_resource_group(group_name)
            self.group_id = resp.group_id
        else:
            if group_id is None:
                raise ValueError(
                    'If create is False you need to provide a valid group_id')

    def add_resource(self, resource):
        '''Add a skil.resource.base.Resource to this group
        '''
        self.skil.api.add_resource_to_group(
            self.group_id, resource.resource_id)

    def delete_resource(self, resource):
        '''Delete a skil.resource.base.Resource from this group
        '''
        self.skil.api.delete_resource_from_group(
            self.group_id, resource.resource_id)

    def delete(self):
        '''Delete this resource group.
        '''
        self.skil.api.delete_resource_group_by_id(self.group_id)

    def get_all_resources(self):
        '''Get all resources attached to this group
        '''
        resp = self.skil.api.get_resources_from_group(self.group_id)
        return [get_resource_by_id(self.skil, res.resource_id) for res in resp]


def delete_resource_group_by_id(skil, group_id):
    skil.api.delete_resource_group_by_id(group_id)


def get_resource_group_by_id(skil, group_id):
    resp = skil.api.get_resource_group_by_id(group_id)
    return ResourceGroup(skil, resp.group_name, resp.group_id, False)


def get_all_resource_groups(skil):
    resp = skil.api.get_resource_groups()
    return [get_resource_group_by_id(skil, group.group_id) for group in resp]


def get_resources_from_group(skil, group):
    resp = skil.api.get_resources_from_group(group.group_id)
    return [get_resource_by_id(skil, res.resource_id) for res in resp]



#!/usr/bin python
# -*- coding: utf-8 -*-
################################################################################
# Copyright (c) 2015-2018 Skymind, Inc.
#
# This program and the accompanying materials are made available under the
# terms of the Apache License, Version 2.0 which is available at
# https://www.apache.org/licenses/LICENSE-2.0.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# SPDX-License-Identifier: Apache-2.0
################################################################################

import argparse
import json
import os
import sys
import pkg_resources
import argcomplete
import traceback
import subprocess
import click
from click.exceptions import ClickException
from dateutil import parser

from skil.experiments import Experiment
from skil.workspaces import WorkSpace
from skil.base import Skil
from skil.deployments import Deployment

from .config import DEFAULT_SKIL_CONFIG, save_skil_config

if sys.version_info[0] == 2:
    input = raw_input


def to_bool(string):
    if type(string) is bool:
        return string
    return True if string[0] in ["Y", "y"] else False


class CLI(object):

    def __init__(self):
        self.var_args = None
        self.command = None

        self.default_host = DEFAULT_SKIL_CONFIG['host']
        self.default_port = DEFAULT_SKIL_CONFIG['port']
        self.default_username = DEFAULT_SKIL_CONFIG['username']
        self.default_password = DEFAULT_SKIL_CONFIG['password']

    def command_dispatcher(self, args=None):
        desc = (
            'Pyskil - train, deploy and manage deep learning experiments with SKIL from Python.\n')
        parser = argparse.ArgumentParser(description=desc)
        parser.add_argument(
            '-v', '--version', action='version',
            version=pkg_resources.get_distribution("skil").version,
            help='Print pyskil version'
        )

        subparsers = parser.add_subparsers(title='subcommands', dest='command')
        subparsers.add_parser(
            'configure', help='Base configuration for pyskil. Run once')

        exp_parser = subparsers.add_parser(
            'init-experiment', help='Initialize a SKIL experiment from scratch.')
        exp_parser.add_argument(
            '-f', '--file', help='File to persist the experiment to.')

        dep_parser = subparsers.add_parser(
            'init-deployment', help='Initialize a SKIL deployment from scratch.')
        dep_parser.add_argument(
            '-f', '--file', help='File to persist the deployment to.')

        argcomplete.autocomplete(parser)
        args = parser.parse_args(args)
        self.var_args = vars(args)

        if not args.command:
            parser.print_help()
            return

        self.command = args.command

        if self.command == 'configure':
            self.configure()
            return

        if self.command == 'init-experiment':
            self.init_experiment(self.var_args['file'])
            return

        if self.command == 'init-deployment':
            self.init_deployment(self.var_args['file'])
            return

    def configure(self):

        click.echo(click.style(u"""\n██████╗ ██╗   ██╗███████╗██╗  ██╗██╗██╗     
██╔══██╗╚██╗ ██╔╝██╔════╝██║ ██╔╝██║██║     
██████╔╝ ╚████╔╝ ███████╗█████╔╝ ██║██║     
██╔═══╝   ╚██╔╝  ╚════██║██╔═██╗ ██║██║     
██║        ██║   ███████║██║  ██╗██║███████╗
╚═╝        ╚═╝   ╚══════╝╚═╝  ╚═╝╚═╝╚══════╝ \n""", fg='blue', bold=True))

        click.echo(click.style("pyskil", bold=True) +
                   " train, deploy and manage deep learning experiments with SKIL from Python!\n")

        host = input("Specify your SKIL host address (default '%s'): " %
                     self.default_host) or self.default_host

        port = input("Specify your SKIL port (default '%s'): " %
                     self.default_port) or self.default_port

        username = input("Specify your SKIL user name (default '%s'): " %
                         self.default_username) or self.default_username

        password = input("Specify your SKIL password (default '%s'): " %
                         self.default_username) or self.default_username
        cli_out = {
            'host': host,
            'port': port,
            'username': username,
            'password': password
        }

        formatted_json = json.dumps(cli_out, sort_keys=False, indent=4)

        click.echo("\nThis is your current settings file " +
                   click.style(".skil", bold=True) + ":\n")
        click.echo(click.style(formatted_json, fg="green", bold=True))

        confirm = input(
            "\nDoes this look good? (default 'y') [y/n]: ") or 'yes'
        if not to_bool(confirm):
            click.echo(
                "" + click.style("Please initialize pyskil once again", fg="red", bold=True))
            return

        save_skil_config(cli_out)

    def init_experiment(self, file_name):
        if not file_name:
            file_name = 'experiment.json'
        path = os.path.join(os.getcwd(), file_name)
        if not os.path.isfile(path):
            experiment = Experiment()
            experiment.save(path)
        else:
            print('Warning: experiment file {} already exists'.format(file_name))

    def init_deployment(self, file_name):
        if not file_name:
            file_name = 'deployment.json'
        path = os.path.join(os.getcwd(), file_name)
        if not os.path.isfile(path):
            deployment = Deployment()
            deployment.save(path)
        else:
            print('Warning: deployment file {} already exists'.format(file_name))


def handle():
    try:
        cli = CLI()
        sys.exit(cli.command_dispatcher())
    except KeyboardInterrupt:
        sys.exit()
    except Exception as e:
        click.echo(click.style("Error: ", fg='red', bold=True))
        traceback.print_exc(e)
        sys.exit()


if __name__ == '__main__':
    handle()



from .base import Resource
from .compute import *
from .storage import *


def get_all_resources(skil):
    '''Get all current SKIL resources as a list of
    skil.resources.base.Resource instances.
    '''
    resource_list = skil.api.get_resources()
    resources = []

    for resource_obj in resource_list:
        resource = Resource(skil)
        resource.resource_id = resource_obj.resource_id
        resources.append(resource)
    return resources


def get_resource_by_id(skil, resource_id):
    '''Get a skil.resources.base.Resource object
    by ID.
    '''
    api_response = skil.api.get_resource_by_id(resource_id)
    return Resource(skil, api_response.resource_id)


def get_resources_by_type(skil, resource_type):
    '''Get a list of skil.resources.base.Resource objects
    by type ('compute' or 'storage').
    '''
    resource_list = skil.api.get_resource_by_type(resource_type)
    resource_ids = [resource.resource_id for resource in resource_list]
    return [get_resource_by_id(skil, resource_id)
            for resource_id in resource_ids]


def get_resources_by_sub_type(skil, sub_type):
    '''Get a list of resources by string sub_type, namely
        - EMR                   # AWS Elastic Map Reduce(Compute)
        - S3                    # AWS Simple Storage Service
        - GoogleStorage         # Google Cloud Storage
        - DataProc              # Google Big Data Compute Engine
        - HDInsight             # Azure Compute
        - AzureStorage          # Azure Blob Storage
        - HDFS                  # in house Hadoop (Storage)

    For instance, choosing 'EMR' sub_type, you'll get all
    skil.resources.compute.EMR resource instances in a Python list.
    '''
    subtype_list = skil.api.get_resource_by_sub_type(sub_type)
    return [get_resource_details_by_id(
        skil, obj.resource_id) for obj in subtype_list]


def get_resource_details_by_id(skil, resource_id):
    '''Get a concrete resource implementation of
    skil.resources.base.Resource by resource ID. For instance, if
    your resource ID corresponds to a resource of subtype "HDFS",
    this will return a skil.resources.storage.HDFS object.
    '''
    resource = skil.api.get_resource_by_id(resource_id)
    res_type = resource.sub_type

    details = skil.api.get_resource_details_by_id(resource_id)
    if res_type == 'EMR':
        return EMR(skil, resource.name, details['region'], None,
                   details['clusterId'], resource_id, False)
    elif res_type == 'S3':
        return S3(skil, resource.name,
                  details['bucket'], details['region'], None, resource_id, False)
    else:
        if res_type == 'GoogleStorage':
            return GoogleStorage(
                skil, resource.name, details['projectId'], details['bucketName'], None, resource_id, False)
        elif res_type == 'DataProc':
            return DataProc(skil, resource.name, details['projectId'], details['region'],
                            details['sparkClusterName'], None, resource_id, False)
        elif res_type == 'HDInsight':
            return HDInsight(skil, resource.name, details['subscriptionId'], details['resourceGroupName'],
                             details['clusterName'], None, resource_id, False)
        elif res_type == 'AzureStorage':
            return AzureStorage(
                skil, resource.name, details['containerName'], None, resource_id, False)
        elif res_type == 'HDFS':
            return HDFS(skil, resource.name, details['nameNodeHost'],
                        details['nameNodePort'], None, resource_id, False)
        elif res_type == 'YARN':
            return YARN(skil, resource.name,
                        details['localSparkHome'], None, resource_id, False)



import json
import os


def mkdir(x):
    if not os.path.isdir(x):
        os.mkdir(x)


_USER_PATH = os.path.expanduser('~')
_DL4J_DIR = os.path.join(_USER_PATH, '.deeplearning4j')
mkdir(_DL4J_DIR)
_BASE_DIR = os.path.join(_DL4J_DIR, 'pyskil')
mkdir(_BASE_DIR)

_SKIL_DIR = os.path.join(_BASE_DIR, '.skil')

DEFAULT_SKIL_CONFIG = {
    'host': 'localhost',
    'port': '9008',
    'username': 'admin',
    'password': 'admin'
}

SKIL_CONFIG = {}


def save_skil_config(config):
    global SKIL_CONFIG
    SKIL_CONFIG = config
    with open(_SKIL_DIR, 'w') as f:
        json.dump(SKIL_CONFIG, f)


def load_skil_config():
    global SKIL_CONFIG
    if os.path.isfile(_SKIL_DIR):
        with open(_SKIL_DIR, 'r') as f:
            SKIL_CONFIG = json.load(f)
    else:
        SKIL_CONFIG = DEFAULT_SKIL_CONFIG



from skil.base import *
from skil.deployments import *
from skil.experiments import *
from skil.models import *
from skil.workspaces import *
from skil.context import *
from skil.services import *
from skil.resources import *
from skil.jobs import *
from skil.utils import *
from skil.config import load_skil_config

# Try to find SKIL config when first importing skil
load_skil_config()



import os
import shutil
import sys
import uuid

import keras
from six import u as unicode
from six import string_types


class SkilContext(object):
    """SkilContext manages models from within a Zeppelin notebook
    hosted by SKIL. The Spark context necessary for initialization
    is provided in any notebook created with SKIL.

    SkilContext can upload models, add them to an experiment and
    add evaluation metrics to a model.  
    """

    def __init__(self, sc):
        self._sc = sc

        self.JvmSkilContext = sc._jvm.io.skymind.zeppelin.utils.SkilContext
        self._ctx = self.JvmSkilContext()

        self.SKILEnvironment = sc._jvm.io.skymind.skil.service.SKILEnvironment
        self.ModelInstanceEntity = sc._jvm.io.skymind.modelproviders.history.model.ModelInstanceEntity
        self.Nd4j = sc._jvm.org.nd4j.linalg.factory.Nd4j
        self.Evaluation = sc._jvm.org.deeplearning4j.eval.Evaluation

    def _models_path(self):
        service_path = self.SKILEnvironment.skilServiceWorkingDirFile().toString()
        storage_path = os.path.join(service_path, "storage")
        models_path = os.path.join(storage_path, "models")

        if not os.path.exists(storage_path):
            os.mkdir(storage_path)
        if not os.path.exists(models_path):
            os.mkdir(models_path)

        return models_path

    def experiment_id(self, z):
        """Get the experiment ID of the current notebook or None.

        # Arguments:
            z: the ZeppelinContext

        # Return value:
            The experiment ID
        """

        return self._ctx.experimentId(z.z)

    def save_model(self, z, model):
        """Save the model into the managed models directory.

        # Arguments:
            z: The ZeppelinContext
            model: The model to save

        # Return value:
            The path of the saved model.
        """

        experiment_id = self.experiment_id(z)
        if experiment_id is None:
            return None

        if isinstance(model, keras.models.Model):
            models_path = self._models_path()
            file_path = os.path.join(models_path, str(uuid.uuid1()) + '.h5')

            if isinstance(file_path, unicode):
                file_path = file_path.encode(sys.getfilesystemencoding())

            model.save(str(file_path))

            return file_path
        else:
            raise NotImplementedError("Only Keras models currently supported.")

    def copy_model(self, z, source_path, model_type):
        """Copy a model file (tensorflow or ONNX) to the managed model directory.

        # Arguments:
            z: The ZeppelinContext
            path: The path to the model you want to copy
            model_type: The type of model. Currently either 'tensorflow' or 'onnx'

        # Return value:
            The path of the saved model
        """

        models_path = self._models_path()

        if model_type.lower() == 'tensorflow' or model_type.lower() == 'tf':
            dest_path = os.path.join(models_path, str(uuid.uuid1()) + '.pb')
        elif model_type.lower() == 'onnx':
            dest_path = os.path.join(models_path, str(uuid.uuid1()) + '.onnx')
        else:
            raise NotImplementedError(
                'Only TensorFlow and ONNX model types are supported.')

        shutil.copyfile(source_path, dest_path)

        return dest_path

    def add_model_to_experiment(self, z, model, name=None):
        """Add the model to the model list inside the SKIL Experiment.

        # Arguments:
            z: The ZeppelinContext
            model: The model to save or path.
            name: Optional name of the model.

        # Return value:
            The ModelInstanceID for use with adding EvaluationResults.
        """

        model_id = str(uuid.uuid1())
        if isinstance(model, string_types):
            model_path = model
        elif isinstance(model, keras.models.Model):
            model_path = self.save_model(z, model)
        else:
            raise NotImplementedError(
                "Can only auto-save Keras models. For TensorFlow or ONNX models please save them separately.")

        instance = (self.ModelInstanceEntity
                    .builder()
                    .modelName(name)
                    .experimentId(self.experiment_id(z))
                    .modelId(model_id)
                    .created(self._sc._jvm.java.util.Date())
                    .uri('file://' + model_path)
                    .etlJson(None)
                    .notebookJson(None)
                    .build())

        self._ctx.getClient().addModelInstance(instance)

        return model_id

    def add_evaluation_to_model(self, z, model_id, model, data, labels, name=None):
        """Gathers the evalutation results of the model on the specified test data.

        # Arguments:
            z: The ZeppelinContext
            model_id: The ModelInstanceID
            model: The model to use
            data: The dataset to evaluate
            labels: The labels of the dataset
            name: A name for the EvaluationResults

        # Return value:
            The id of the EvaluationResults if saved.
        """

        def assign(np_arr, jvm_arr):
            i = 0
            for row in np_arr.tolist():
                j = 0
                for col in row:
                    jvm_arr[i][j] = col
                    j += 1
                i += 1

        if not isinstance(model, keras.models.Model):
            raise NotImplementedError("Only Keras models currently supported.")

        preds = model.predict(data)
        p_arr = self._sc._gateway.new_array(
            self._sc._gateway.jvm.double, preds.shape[0], preds.shape[1])
        assign(preds, p_arr)
        nd_pred = self.Nd4j.create(p_arr)
        y_arr = self._sc._gateway.new_array(
            self._sc._gateway.jvm.double, labels.shape[0], labels.shape[1])
        assign(labels, y_arr)
        nd_y = self.Nd4j.create(y_arr)

        e = self.Evaluation(int(nd_y.shape()[1]))
        e.eval(nd_y, nd_pred)

        return self._ctx.addEvaluationToModel(z.z, model_id, e, name)

    # Camel case methods for backward compatibility
    def experimentId(self, z):
        return self.experiment_id(z)

    def saveModel(self, z, model):
        return self.save_model(z, model)

    def copyModel(self, z, source_path, model_type):
        return self.copy_model(z, source_path, model_type)

    def addModelToExperiment(self, z, model, name=None):
        return self.add_model_to_experiment(z, model, name)

    def addEvaluationToModel(self, z, model_id, model, data, labels, name=None):
        return self.add_evaluation_to_model(z, model_id, model, data, labels, name)



from setuptools import setup
from setuptools import find_packages

setup(
    name='skil',
    version='1.2.0',
    packages=find_packages(),
    install_requires=[
        'skil_client>=1.2.1', 
        'requests', 
        'numpy', 
        'click', 
        'argcomplete',
        'tensorflow', 
        'keras', 
        'opencv-python', 
        'six'],
    extras_require={
        'aws': ['aws'],
        'azure': ['azure'],
        'gce': ['google-api-python-client'],
        'tests': ['pytest', 'pytest-pep8', 'pytest-cov', 'mock']
    },
    include_package_data=True,
    license='Apache',
    description='Train, deploy and manage your Python models with SKIL',
    long_description='Train, deploy and manage your Python models with SKIL',
    author='Max Pumperla',
    author_email='max@skymind.io',
    url='https://github.com/SkymindIO/skil-python',
    entry_points={
        'console_scripts': [
            'pyskil=skil.cli:handle'
        ]
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Environment :: Console',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3'
    ]
)



import skil_client
from skil_client.rest import ApiException as api_exception

try:
    from skil_client import Credentials as Cred
except:
    from skil_client import LoginRequest as Cred

import pprint
import os
import time
import requests
import json
import subprocess
from .config import SKIL_CONFIG, save_skil_config


class Skil:
    def __init__(self, workspace_server_id=None, host='localhost', port=9008,
                 debug=False, user_id='admin', password='admin'):

        self.printer = pprint.PrettyPrinter(indent=4)

        config = skil_client.Configuration()
        config.host = "{}:{}".format(host, port)
        config.debug = debug
        self.config = config
        self.uploads = []
        self.uploaded_model_names = []
        self.auth_headers = None

        self.api_client = skil_client.ApiClient(configuration=config)
        self.api = skil_client.DefaultApi(api_client=self.api_client)

        try:
            self.printer.pprint('>>> Authenticating SKIL...')
            credentials = Cred(user_id=user_id, password=password)

            token = self.api.login(credentials)
            self.token = token.token
            config.api_key['authorization'] = self.token
            config.api_key_prefix['authorization'] = "Bearer"
            self.printer.pprint('>>> Done!')
        except api_exception as e:
            raise Exception(
                "Exception when calling DefaultApi->login: {}\n".format(e))

        if workspace_server_id:
            self.server_id = workspace_server_id
        else:
            self.server_id = self.get_default_server_id()

        # Store config for future connections
        base_config = {
            'host': host,
            'port': port,
            'username': user_id,
            'password': password
        }
        save_skil_config(base_config)

    @classmethod
    def from_config(cls):
        return Skil(**SKIL_CONFIG)

    def get_default_server_id(self):
        self.auth_headers = {'Authorization': 'Bearer %s' % self.token}
        r = requests.get(
            'http://{}/services'.format(self.config.host), headers=self.auth_headers)
        if r.status_code != 200:
            r.raise_for_status()

        content = json.loads(r.content.decode('utf-8'))
        services = content.get('serviceInfoList')
        server_id = None
        for s in services:
            if 'Model History' in s.get('name'):
                server_id = s.get('id')
        if server_id:
            return server_id
        else:
            raise Exception(
                "Could not detect default model history server instance. Is SKIL running?")

    def upload_model(self, model_name):
        self.printer.pprint('>>> Uploading model, this might take a while...')
        upload = self.api.upload(file=model_name).file_upload_response_list
        self.uploads = self.uploads + upload
        self.uploaded_model_names.append(model_name)
        self.printer.pprint(self.uploads)

    def get_uploaded_model_names(self):
        return self.uploaded_model_names

    def get_model_path(self, model_name):
        for upload in self.uploads:
            if model_name == upload.file_name:
                return "file://" + upload.path
        raise Exception("Model resource not found, did you upload it? ")
